// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import java.util.HashMap;
import java.util.Map;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.updatesimplifier.Update;
import de.uka.ilkd.key.util.Debug;


/**
 * Creates anonymising updates for sets of locations, i.e. (quantified) updates
 * which assign uninterpreted functions to all locations of such a set.
 */
public class AnonymisingUpdateFactory {
    private static final TermBuilder TB = TermBuilder.DF;
    private static final TermFactory TF = TermFactory.DEFAULT;
    private final UpdateFactory uf;
    
    
    public AnonymisingUpdateFactory(UpdateFactory uf) {
        this.uf = uf;
    }

    
    private static Term[] getSubTerms(Term term) {
        int arity = term.arity();
        Term[] result = new Term[arity];
        
        for(int i = 0; i < arity; i++) {
            result[i] = term.sub(i);
        }
        
        return result;
    }
    
    
    private static Name getNewName(Services services, Name baseName) {
        NamespaceSet namespaces = services.getNamespaces();
        
        int i = 0;
        Name name;
        do {
            name = new Name(baseName + "_" + i++);
        } while(namespaces.lookup(name) != null);
        
        return name;
    }
       
    
    /**
     * Determines the arguments sorts of the passed operator
     * (helper for getUninterpretedFunction()).
     */
    private static Sort[] getArgumentSorts(Operator op, Services services) {
        Sort[] result = new Sort[op.arity()];
        
        if(op instanceof ProgramVariable) {
            Debug.assertTrue(op.arity() == 0);
        } else if(op instanceof AccessOp) {
            Sort integerSort = services.getTypeConverter().getIntegerLDT()
                                                          .targetSort();

            if(op instanceof AttributeOp) {
                AttributeOp aop = (AttributeOp) op;
                KeYJavaType kjt = ((ProgramVariable) aop.attribute())
                                  .getContainerType();
                if(services.getJavaInfo().rec2key().getSuperArrayType() == kjt
                   && "length".equals(aop.attribute().name().toString())) {
                    result[0] 
                         = services.getJavaInfo().getJavaLangObjectAsSort();
                } else {
                    result[0] = kjt.getSort();
                }
            } else if(op instanceof ArrayOp) {
                result[0] = ((ArrayOp)op).arraySort();
                result[1] = integerSort;
            } else {
                Debug.fail("Unexpected location operator."+op);
            }

            if(((AccessOp)op).isShadowed()) {
                result[result.length - 1] = integerSort; 
            }
        } else if(op instanceof Function) {           
            Function func = (Function) op;               
            for(int i = 0; i < result.length; i++) {
                result[i] = func.argSort(i);
            }
        } else {
            Debug.fail("Unsupported location operator."+op);  
        }
        
        return result;
    }
    
    
    /**
     * Gets an uninterpreted function for the passed location term 
     * (helper for createUninterpretedFunctions()).
     */
    private static RigidFunction getUninterpretedFunction(
                                Term locTerm, 
                                Map /*Operator -> RigidFunction*/ functions,
                                Services services) {
        RigidFunction result = (RigidFunction) functions.get(locTerm.op());
        
        if (result == null) {
            Name baseName = locTerm.op() instanceof AttributeOp ? 
                    ((AttributeOp) locTerm.op()).attribute().name()
                    : locTerm.op() instanceof ArrayOp ? new Name("get")
                            : locTerm.op().name();

            if (baseName instanceof ProgramElementName) {
                baseName = new 
                    Name(((ProgramElementName)baseName).getProgramName());
            }

            result = new RigidFunction(getNewName(services, baseName), locTerm
                    .sort(), getArgumentSorts(locTerm.op(), services));
            services.getNamespaces().functions().add(result);
            functions.put(locTerm.op(), result);
        }
        
        return result;
    }
    
    
    /**
     * Creates suitable uninterpreted functions for the passed locations.
     */
    public static RigidFunction[] createUninterpretedFunctions(
                                            LocationDescriptor[] locations,
                                            Services services) {
        RigidFunction[] result = new RigidFunction[locations.length];
        Map /*Operator -> RigidFunction*/ functions = new HashMap();
        
        for(int i = 0; i < locations.length; i++) {
            if(locations[i] instanceof BasicLocationDescriptor) {
                BasicLocationDescriptor bloc 
                        = (BasicLocationDescriptor) locations[i];
                result[i] = getUninterpretedFunction(bloc.getLocTerm(), 
                                                     functions,
                                                     services);
            } else {
                Debug.assertTrue(
                        locations[i] instanceof EverythingLocationDescriptor);
                result[i] = null;
            }
        }
        
        return result;
    }
    
    
    /**
     * Creates the anonymising update for the passed locations using the passed
     * uninterpreted functions (which must have the correct signature).
     */
    public Update createAnonymisingUpdate(LocationDescriptor[] locations,
                                          Function[] functions,
				          Services services) {
        Debug.assertTrue(functions.length == locations.length);
        
        Update result = uf.skip();

        for(int i = 0; i < locations.length; i++) {
            if(locations[i] instanceof BasicLocationDescriptor) {
                BasicLocationDescriptor bloc 
                        = (BasicLocationDescriptor) locations[i];
                Term guardTerm = bloc.getFormula();
                Term locTerm = bloc.getLocTerm();

                //optimisation: for location descriptors of the form
                //"\for int x; \if(0 <= x & x < a.length) a[x]" 
                //(resulting e.g. from parsing "a[*]"),
                //we skip the guard in order to simplify the result.
                if(locTerm.op() instanceof ArrayOp 
                   && locTerm.sub(1).op() instanceof LogicVariable) {
		    ProgramVariable lengthPV 
			= services.getJavaInfo().getArrayLength();
		    Function sub 
			= services.getTypeConverter().getIntegerLDT()
			    	     	             .getArithSubstraction();

                    Term varTerm 
			= TB.var((LogicVariable)locTerm.sub(1).op());
		    Term lengthTerm 
			= TB.dot(locTerm.sub(0), lengthPV);
		    Term lengthMinusOneTerm 
			= TB.func(sub, lengthTerm, TB.one(services));

		    Term lowerBoundTerm 
			= TB.leq(TB.zero(services), varTerm, services);
		    Term upperBoundTerm1 
			= TB.lt(varTerm, lengthTerm, services);
		    Term upperBoundTerm2 
			= TB.leq(varTerm, lengthMinusOneTerm, services);
		    if(guardTerm.equals(TB.and(lowerBoundTerm, 
					       upperBoundTerm1))
		       || guardTerm.equals(TB.and(lowerBoundTerm, 
						  upperBoundTerm2))) {
			guardTerm = TB.tt();
		    }
                }
                
                //create elementary update
                Term[] argTerms = getSubTerms(locTerm);
                Term valueTerm = TF.createFunctionTerm(functions[i], argTerms);
                Update elementaryUpdate = uf.elementaryUpdate(locTerm, 
                                                              valueTerm);
                
                //create guarded update
                Update guardedUpdate = uf.guard(guardTerm, elementaryUpdate);
                
                //create quantified update
                Update quantifiedUpdate = guardedUpdate;
                IteratorOfQuantifiableVariable it 
                        = locTerm.freeVars().iterator();
                while(it.hasNext()) {
                    quantifiedUpdate = uf.quantify(it.next(), quantifiedUpdate);
                }
                //add update to result update
                result = uf.parallel(result, quantifiedUpdate);                
            } else {
                Debug.assertTrue(
                        locations[i] instanceof EverythingLocationDescriptor);
                
                //create anonym*ous* update
                AnonymousUpdate ao = AnonymousUpdate.getNewAnonymousOperator();
                Term[] trueTerm = new Term[] {TF.createJunctorTerm(Op.TRUE)};
                Term updateTerm = TF.createAnonymousUpdateTerm(ao, trueTerm);
                return Update.createUpdate(updateTerm);
            }
        }

        return result;
    }
    
    
    /**
     * Creates the anonymising update for the passed locations using new 
     * uninterpreted functions.
     */
    public Update createAnonymisingUpdate(SetOfLocationDescriptor locations,
                                          Services services) {
        LocationDescriptor[] locationsArray = locations.toArray();
        RigidFunction[] functions = createUninterpretedFunctions(locationsArray, 
                                                                 services);
        return createAnonymisingUpdate(locationsArray, functions, services);
    }
    
    
    /**
     * Creates the anonymising update for the passed locations using the passed
     * uninterpreted functions (which must have the correct signature) and 
     * applies it to the passed target term.
     */
    public Term createAnonymisingUpdateTerm(LocationDescriptor[] locations,
                                            Function[] functions,
                                            Term targetTerm,
					    Services services) {
        return uf.apply(createAnonymisingUpdate(locations, functions, services),
                        targetTerm);
    }
    
    
    /**
     * Creates the anonymising update for the passed locations using new
     * uninterpreted functions and applies it to the passed target term.
     */
    public Term createAnonymisingUpdateTerm(SetOfLocationDescriptor locations,
                                            Term targetTerm,
                                            Services services) {
        LocationDescriptor[] locationsArray = locations.toArray();
        RigidFunction[] functions = createUninterpretedFunctions(locationsArray,
                                                                 services);
        return createAnonymisingUpdateTerm(locationsArray, 
				           functions, 
					   targetTerm, 
					   services);
    }
}
