// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

/**
 * Implements "outer renaming", i.e. renaming the old variable in case the
 * name of a new variable entering the globals clashes with it.
 */

package de.uka.ilkd.key.logic;

import java.util.HashMap;

import de.uka.ilkd.key.java.ArrayOfProgramElement;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ProgVarReplacer;
import de.uka.ilkd.key.proof.RuleAppIndex;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.rule.IteratorOfNoPosTacletApp;
import de.uka.ilkd.key.rule.ListOfNoPosTacletApp;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.util.Debug;


public class OuterVariableNamer extends VariableNamer {

    public OuterVariableNamer(Services services) {
    	super(services);
    }

    
    /**
     * returns the global variable which has the same name as the passed
     * variable, or null if no such variable exists
     */
    private ProgramVariable getConflictingVar(ProgramVariable var, 
                                              ProgramElementName name, 
					      Goal goal) {
        IteratorOfProgramVariable it = goal.getGlobalProgVars().iterator();
        while(it.hasNext()) {
	    ProgramVariable var2 = it.next();
            ProgramElementName name2 = var2.getProgramElementName();
            if(var2 != var && name2.equals(name)) {
            	return var2;
            }
        }

        return null;
    }
    
    
    /**
     * returns the maximum counter existing for the passed basename in all 
     * programs contained in the passed SVInstantiations, or -1
     * (helper for getMaxCounterInTacletApps())
     */
    private int getMaxCounterInInsts(String basename, SVInstantiations insts) {
    	int result = -1;
    
    	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it;
	it = insts.pairIterator();
	while(it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry e = it.next();
	    InstantiationEntry ie = e.value();
	    Object inst = ie.getInstantiation();

	    if(ie instanceof ContextInstantiationEntry) {
	    	int maxInProgram = getMaxCounterInProgram(basename,
							  (ProgramElement)inst,
							  null);
		if(maxInProgram > result) {
		    result = maxInProgram;
		}
	    } else if(ie instanceof OperatorInstantiation) {
		/*nothing to be done (currently)*/
	    } else if(ie instanceof ProgramInstantiation) {
	    	int maxInProgram = getMaxCounterInProgram(basename, 
							  (ProgramElement)inst,
							  null);
		if(maxInProgram > result) {
		    result = maxInProgram;
		}
	    } else if(ie instanceof ProgramListInstantiation) {
		ArrayOfProgramElement a = (ArrayOfProgramElement) inst;
		int size = a.size();
		for(int i = 0; i < size; i++) {
	    	    int maxInProgram 
		    	= getMaxCounterInProgram(basename, 
					  	 a.getProgramElement(i),
						 null);
		    if(maxInProgram > result) {
		    	result = maxInProgram;
		    }		    
		}
	    } else if(ie instanceof TermInstantiation) {
	    	int maxInTerm = getMaxCounterInTerm(basename, (Term)inst);
		if(maxInTerm > result) {
		    result = maxInTerm;
		}
	    } else {
		Debug.fail("unexpected subtype of InstantiationEntry");
	    }
	}
	
	return result;
    }


    /**
     * returns the maximum counter existing for the passed basename in the
     * passed goal's taclet apps, or -1
     * (helper for getMaxCounterInGlobalsAndAllPrograms())
     */
    private int getMaxCounterInTacletApps(String basename, Goal goal) {
	RuleAppIndex ruleAppIndex = goal.ruleAppIndex();
	TacletIndex tacletIndex = ruleAppIndex.tacletIndex();
	ListOfNoPosTacletApp noPosTacletApps
		= tacletIndex.getPartialInstantiatedApps();

	int result = -1;
	IteratorOfNoPosTacletApp it = noPosTacletApps.iterator();
	while(it.hasNext()) {
	    SVInstantiations insts = it.next().instantiations();
	    int maxInInsts = getMaxCounterInInsts(basename, insts);
	    if(maxInInsts > result) {
	    	result = maxInInsts;
	    }
	}
	
	return result;
    }

    
    /**
     * returns the maximum counter existing for the passed basename in the 
     * passed term, or -1
     * (helper for getMaxCounterInGlobalsAndAllPrograms())
     */
    private int getMaxCounterInTerm(String basename, Term term) {
    	JavaBlock jb = term.javaBlock();
	if(!jb.isEmpty()) {
	    return getMaxCounterInProgram(basename, jb.program(), null);
	}

	int result = -1;
    	int arity = term.arity();
	for(int i = 0; i < arity; i++) {
	    int maxInSubterm = getMaxCounterInTerm(basename, term.sub(i));
	    if(maxInSubterm > result) {
	    	result = maxInSubterm;
	    }
	}

    	return result;
    }


    /**
     * returns the maximum counter for the passed basename anywhere in the
     * passed goal
     */
    private int getMaxCounterInGlobalsAndAllPrograms(String basename,
    						     Goal goal) {
	int maxInGlobals
	       = getMaxCounterInGlobals(basename,
					wrapGlobals(goal.getGlobalProgVars()));
					
	int maxInPrograms = getMaxCounterInTacletApps(basename, goal);
    
    	IteratorOfConstrainedFormula it = goal.sequent().iterator();
	while(it.hasNext()) {
	    Term term = it.next().formula();
	    int maxInTerm = getMaxCounterInTerm(basename, term);
	    if(maxInTerm > maxInPrograms) {
	    	maxInPrograms = maxInTerm;
	    }
	}

	return (maxInGlobals > maxInPrograms ? maxInGlobals : maxInPrograms);
    }


    public ProgramVariable rename(ProgramVariable var,
                                  Goal goal,
                                  PosInOccurrence posOfFind) {
	ProgramElementName name = var.getProgramElementName();
	BasenameAndIndex bai = getBasenameAndIndex(name);
	Globals globals = wrapGlobals(goal.getGlobalProgVars());
	map = new HashMap();
        renamingHistory = new HashMap();
	//prepare renaming of inner var
	ProgramElementName newname = createName(bai.basename,
	 				        bai.index, 
						getMethodStack(posOfFind));
	ProgramVariable newvar 
		= new LocationVariable(newname, var.getKeYJavaType());
	map.put(var, newvar);
        renamingHistory.put(var, newvar);
	//prepare renaming of outer var, if existent
	ProgramVariable confVar = getConflictingVar(var, newname, goal);
	if(confVar != null) {
	    int counter = getMaxCounterInGlobalsAndAllPrograms(bai.basename,
	    						       goal) + 1;
	    Debug.assertTrue(counter != 0);
	    ProgramElementName newConfName
	    	= createName(bai.basename,
			     counter,
	  		     confVar.getProgramElementName().getCreationInfo());
	    ProgramVariable newConfVar
	    	= new LocationVariable(newConfName, confVar.getKeYJavaType());
	    map.put(confVar, newConfVar);
	    if (confVar.equals(newvar)){
                renamingHistory.clear();
                renamingHistory.put(var,newConfVar);
                
            } else {
                renamingHistory.put(confVar,newConfVar);
                
            }
	}

	//execute renaming
	ProgVarReplacer pvr = new ProgVarReplacer(map);
	pvr.replace(goal);

	return newvar;
    }
}
