package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.ArrayOfLocationDescriptor;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.util.Debug;


/**
 * represents a non-rigid function whose value depend on the value
 * of locations given in form of location descriptors. This function shall
 * replace the former @link {@link NRFunctionWithExplicitDependencies} as
 * it is more flexible
 *
 */
public class LocationDependentFunction extends NonRigidFunction {

    /** the kind of the symbol */
    private final Name kind;
    
    /** 
     * the location descriptor characterising the set of locations the 
     * symbols value depends on
     */
    private final ArrayOfLocationDescriptor locDescr;

    
    /**
     * creates a location dependent function of the given <code>kind</code> whose
     * value depend only on the locations which are described by the
     * given @link LocationDescriptor 
     * @param kind the Name describing the kind of the function
     * @param sort the Sort the function is being mapped to
     * @param argSorts an array of Sorts of the arguments
     * @param locDescr the LocationDescriptor with all locations on which the
     * function's value depends
     */
    public LocationDependentFunction(Name kind, Sort sort, Sort[] argSorts,
            ArrayOfLocationDescriptor locDescr) {
        super(kind, sort, argSorts);
        this.kind     = kind;
        this.locDescr = locDescr;
    }
    
    public MatchConditions match(SVSubstitute subst, MatchConditions mc, Services services) {
        if (this == subst) { 
            return mc;
        }
        if (subst instanceof LocationDependentFunction) {
            final LocationDependentFunction substFunction = (LocationDependentFunction) subst;
            
            if (substFunction.kind().equals(kind())) {
                if (locationDescriptor().size() != substFunction.locationDescriptor().size()) {
                    Debug.out("Match FAILED. Incompatible size of loc descriptors.", this, 
                            substFunction);
                    return null;
                }
                
                MatchConditions result = mc;
                for (int i = 0, sz = locationDescriptor().size(); i<sz; i++) {
                  result = locationDescriptor().getLocationDescriptor(i).match(substFunction.
                          locationDescriptor().getLocationDescriptor(i), result, services);
                  if (result == null) {
                      return null;
                  }
                }
                return result;   
            }    
            Debug.out("Match FAILED. Location dependent functions are of different kind", 
                    this, subst);
            return null;
        }
        Debug.out("Match FAILED. Incompatible function classes", this, subst);
        return null;
    }

    /**
     * returns the kind of this location dependent function
     * @return the kind of this location dependent function
     */
    public Name kind() {
        return kind;
    }
    
    /**
     * the location descriptors characterising the set of locations this function 
     * depends on
     */
    public ArrayOfLocationDescriptor locationDescriptor() {
        return locDescr;
    }
    
    
    /** toString **/ 
    public String toString() {
        return kind + "[" + locDescr + "] ("+ argSort() +"):"+sort();
    }

    
    
}
