// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.java.recoderext.ImplicitFieldAdder;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.Junctor;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.Quantifier;
import de.uka.ilkd.key.logic.sort.AbstractCollectionSort;
import de.uka.ilkd.key.logic.sort.PrimitiveSort;
import de.uka.ilkd.key.util.Debug;


/**
 * Convenience class for creating terms related to the implicit 
 * "created" attribute.
 * Should possibly be merged into TermBuilder?
 */
public class CreatedAttributeTermFactory {
    public static final CreatedAttributeTermFactory INSTANCE 
            = new CreatedAttributeTermFactory();
    
    private static final TermFactory tf = TermFactory.DEFAULT;
    
    private ProgramVariable createdAttribute;
    private Term trueLitTerm;
    
    
    private CreatedAttributeTermFactory() {}
    
    
    private void initialiseMembers(Services services) {
        if(createdAttribute == null) {
            JavaInfo javaInfo = services.getJavaInfo();
            
            createdAttribute
                = javaInfo.getAttribute(ImplicitFieldAdder.IMPLICIT_CREATED, 
                                        javaInfo.getJavaLangObject());
            trueLitTerm = services.getTypeConverter()
                                  .convertToLogicElement(BooleanLiteral.TRUE);
        }
        
        Debug.assertTrue(createdAttribute != null && trueLitTerm != null);
    }
    
    
    /**
     * Creates the formula "objectTerm.<created> = TRUE".
     */
    public Term createCreatedTerm(Services services, Term objectTerm) {
        initialiseMembers(services);
        
        Term createdTerm = tf.createAttributeTerm(createdAttribute, 
                                                  objectTerm);
        
        return tf.createEqualityTerm(createdTerm, trueLitTerm);
    }
    
    
    /**
     * Creates the formula "objectTerm.<created> = TRUE | objectTerm = null"
     */
    public Term createCreatedOrNullTerm(Services services, Term objectTerm) {
        Term objectCreatedTerm = createCreatedTerm(services, objectTerm);
        Term objectNullTerm 
            = tf.createEqualityTerm(objectTerm, 
                                    services.getJavaInfo().getNullTerm());

        return tf.createJunctorTerm(Op.OR, objectCreatedTerm, objectNullTerm);
    }


    /**
     * Creates the formula "objectTerm.<created> = TRUE & objectTerm != null"
     */
    public Term createCreatedAndNotNullTerm(Services services, Term objectTerm) {
        Term objectCreatedTerm = createCreatedTerm(services, objectTerm);
        Term objectNotNullTerm 
            = tf.createEqualityTerm(objectTerm, 
                                    services.getJavaInfo().getNullTerm());
        objectNotNullTerm = tf.createJunctorTerm(Op.NOT, objectNotNullTerm);
        
        return tf.createJunctorTerm(Op.AND, objectCreatedTerm, objectNotNullTerm);
    }


    private Term createQuantifierTerm(Services services,
                                      Quantifier q,
                                      LogicVariable[] vars,
                                      Term subTerm,
                                      boolean nullForbidden) {
        //create conjunction of guard terms for all variables of a
        //non-primitive type
        Term guardConjunctionTerm = tf.createJunctorTerm(Op.TRUE);
        for(int i = 0; i < vars.length; i++) {
            if(!(vars[i].sort() instanceof PrimitiveSort) &&
               !(vars[i].sort() instanceof AbstractCollectionSort)) {
                Term variableTerm = tf.createVariableTerm(vars[i]);
                Term guardTerm
                        = (nullForbidden
                           ? createCreatedAndNotNullTerm(services, variableTerm)
                           : createCreatedTerm(services, variableTerm));
                guardConjunctionTerm 
                       = tf.createJunctorTermAndSimplify(Op.AND,
                                                         guardConjunctionTerm,
                                                         guardTerm);
            }
        }

        //link conjunction and subterm
        Junctor j = (q.equals(Quantifier.ALL) ? Op.IMP : Op.AND);
        Term impTerm = tf.createJunctorTermAndSimplify(j, 
                                                       guardConjunctionTerm, 
                                                       subTerm);
        
        //create quantification
        Term quantifierTerm = tf.createQuantifierTerm(q, vars, impTerm);
            
        return quantifierTerm;
    }

    
    /**
     * Creates a quantifier term where the quantification only covers
     * objects which have already been created.
     */
    public Term createCreatedQuantifierTerm(Services services,
                                            Quantifier q, 
                                            LogicVariable[] vars, 
                                            Term subTerm) {
        return createQuantifierTerm(services, q, vars, subTerm, false);
    }
    
    
    /**
     * Creates a quantifier term where the quantification only covers 
     * objects which have already been created and which are not null.
     */
    public Term createCreatedNotNullQuantifierTerm(Services services,
                                                   Quantifier q, 
                                                   LogicVariable[] vars, 
                                                   Term subTerm) {
        return createQuantifierTerm(services, q, vars, subTerm, true);
    }
    

    /**
     * Creates a quantifier term where the quantification only covers
     * objects which have already been created.
     */
    public Term createCreatedQuantifierTerm(Services services,
                                            Quantifier q,
                                            LogicVariable var,
                                            Term subTerm) {
        return createCreatedQuantifierTerm(services,
                                           q,
                                           new LogicVariable[] {var},
                                           subTerm);
    }


    /**
     * Creates a quantifier term where the quantification only covers 
     * objects which have already been created and which are not null.
     */
    public Term createCreatedNotNullQuantifierTerm(Services services,
	                                           Quantifier q,
			                           LogicVariable var,
			                           Term subTerm) {
        return createCreatedNotNullQuantifierTerm(services,
	                                          q,
				                  new LogicVariable[] {var},
				                  subTerm);
    }
}
