// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import java.util.Iterator;
import java.util.LinkedList;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;

/**
 * Built in rule that starts the update simplifier on the sequent or
 * selected term. The rule is useful if one has changes the simultaneous
 * update simplifier settings. 
 */
public class UpdateSimplificationRule implements BuiltInRule {
    

    // container for old constrained formula and the one used to replace it
    static class ConstrainedFormulaContainer {

	private ConstrainedFormula old;
	private ConstrainedFormula replacement;
	
	public ConstrainedFormulaContainer
	    (ConstrainedFormula old, ConstrainedFormula replacement) {
	    this.old = old;
	    this.replacement = replacement;
	}

	public ConstrainedFormula old() {
	    return old;
	}

	public ConstrainedFormula replacement() {
	    return replacement;
	}

    }

    public static UpdateSimplificationRule INSTANCE = new UpdateSimplificationRule();

    private final Name name = new Name("Update Simplification");

    
    private UpdateSimplificationRule() {
    }

    /** the rule is applied on the given goal using the
     * information of rule application. As applications of this rule kind may
     * not be successful in each case one has to ensure that the goal split is
     * done only iff the application was successful.
     * @param goal the goal that the rule application should refer to.
     * @param services the Services with the necessary information
     * about the java programs 
     * @param the rule application that is executed.
     */
    public ListOfGoal apply(Goal     goal, 
			    Services services, 
			    RuleApp  ruleApp) {

	ListOfGoal result = null;
	LinkedList applicationResult = null;
	
	final PosInOccurrence pio = ruleApp.posInOccurrence();	    	
	final Sequent sequent = goal.sequent();
	if (pio == null) {		
	    applicationResult = applyOnSequent(sequent, 
		    			       goal.simplifier(),
		    			       services);	    
	} else {
	    applicationResult = applyOnTerm(goal.simplifier(), pio, services);
	}	


	if (applicationResult != null && applicationResult.size() >0) { 
	    result = goal.split(1);
	    final Goal newGoal = result.head();
	    final Iterator it = applicationResult.iterator();
	    while (it.hasNext()) {
		ConstrainedFormulaContainer cfc = 
		    (ConstrainedFormulaContainer) it.next();
		
		final PosInOccurrence oldCFPIO = pio == null ? 
		    new PosInOccurrence(cfc.old(), PosInTerm.TOP_LEVEL,	sequent) :
		    new PosInOccurrence(cfc.old(), PosInTerm.TOP_LEVEL,	pio.isInAntec());
		newGoal.changeFormula(cfc.replacement(), oldCFPIO);
	    }
	}
	return result;
    }



    private LinkedList applyOnSequent(Sequent seq, 
				      UpdateSimplifier sus,
				      Services services) {			

	LinkedList result = new LinkedList();

	IteratorOfConstrainedFormula it = seq.iterator();
	Term simplified;
	while (it.hasNext()) {
	    ConstrainedFormula cf = it.next();
	    simplified = sus.simplify(cf.formula(), services);
	    if (simplified != cf.formula()) {
		result.add
		    (new ConstrainedFormulaContainer
			(cf, 
			 new ConstrainedFormula(simplified, cf.constraint()))
			);		
	    }
	}

	return result;
    }

    private LinkedList applyOnTerm(UpdateSimplifier sus,
				   PosInOccurrence pio,
				   Services services) {

	LinkedList result = new LinkedList();
	
	ConstrainedFormula cf = pio.constrainedFormula();
	Term fml = cf.formula();

	Term simplified = sus.simplify(pio.subTerm(), services);
	if (simplified != fml) {
	    ConstrainedFormula new_cf = new ConstrainedFormula
		(replace(fml, simplified, pio.posInTerm().iterator()), 
		 cf.constraint()); 
	    result.add(new ConstrainedFormulaContainer(cf, new_cf));
	}

	return result;
    }

    public String displayName() {
	return toString();
    }
    
    /**
     * tests if the rule is applicable at the given position
     */
    public boolean isApplicable(Goal            goal, 
				PosInOccurrence pio,
				Constraint      userConstraint) {
	Services services = goal.proof().getServices();
	if (pio != null) {
	    Term fml = pio.subTerm ();
	    Term simplified = goal.simplifier().simplify(fml, services);
	    if (!simplified.equals(fml)) {
		return true;
	    }
	} else { // this may be too slow; may be return true always
 	    IteratorOfConstrainedFormula it = goal.sequent().iterator();
 	    Term simplified;
 	    while (it.hasNext()) {
 		ConstrainedFormula cf = it.next();
 		simplified = goal.simplifier().simplify(cf.formula(), services);
 		if (!simplified.equals(cf.formula())) {
 		    return true;
 		}
 	    }
	}
	return false;
    }

    public Name name() {
	return name;
    }


    private Term replace(Term term, Term with, IntIterator it) {	
	if (it.hasNext()) {	    
	    int sub=it.next();
	    Term[] subs=new Term[term.arity()];
	    final ArrayOfQuantifiableVariable[] vars = 
		new ArrayOfQuantifiableVariable[term.arity()];        
	    for (int i=0;i<term.arity();i++) {
		if (i!=sub) {
		    subs[i]=term.sub(i);
		} else {
		    subs[i]=replace(term.sub(i), with, it);
		}
		vars[i] = term.varsBoundHere(i);
	    }
	    
	    return TermFactory.DEFAULT.createTerm(term.op(), subs, vars,
						  term.javaBlock());	    
	}

	return with;
    }

    public String toString() {
	return name().toString();
    }
        
}
