// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.rule.metaconstruct.arith;

import java.math.BigInteger;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.op.AbstractMetaOperator;
import de.uka.ilkd.key.logic.op.Operator;

/**
 * Class for analysing and modifying polynomial expressions over the integers
 */
public class Polynomial {

    private final BigInteger constantPart;
    private final ListOfMonomial parts;

    private Polynomial(ListOfMonomial parts, BigInteger constantPart) {
        this.parts = parts;
        this.constantPart = constantPart;
    }
    
    public static Polynomial create(Term polynomial, Services services) {
        final Analyser a = new Analyser ( services );
        a.analyse ( polynomial );
        return new Polynomial ( a.parts, a.constantPart );
    }

    public Polynomial multiply(BigInteger c) {
        ListOfMonomial newParts = SLListOfMonomial.EMPTY_LIST;
        final IteratorOfMonomial it = parts.iterator ();
        while ( it.hasNext () )
            newParts = newParts.prepend ( it.next ().multiply ( c ) );

        return new Polynomial ( newParts, constantPart.multiply ( c ) );
    }

    /**
     * @return the greatest common divisor of the coefficients of the monomials
     *         of this polynomial. The constant part of the polynomial is not
     *         taken into account. If there are no monomials (apart from the
     *         constant term), the result is <code>BigInteger.ZERO</code>
     */
    public BigInteger coeffGcd() {
        BigInteger res = BigInteger.ZERO;
        final IteratorOfMonomial it = parts.iterator ();
        while ( it.hasNext () )
            res = res.gcd ( it.next ().getCoefficient () );
        return res;
    }
    
    /**
     * @return <code>true</code> if the value of <code>this</code> will
     *         always be less than the value of <code>p</code>
     *         (i.e., same monomials, but the constant part is less or equal)
     */
    public boolean valueLess(Polynomial p) {
        if ( !sameParts ( p ) ) return false;
        return constantPart.compareTo ( p.constantPart ) < 0;
    }

    /**
     * @return <code>true</code> if the value of <code>this</code> will
     *         always be less or equal than the value of <code>p</code>
     *         (i.e., same monomials, but the constant part is less or equal)
     */
    public boolean valueLeq(Polynomial p) {
        if ( !sameParts ( p ) ) return false;
        return constantPart.compareTo ( p.constantPart ) <= 0;
    }

    private boolean sameParts(Polynomial p) {
        if ( parts.size () != p.parts.size () ) return false;
        return difference ( parts, p.parts ).isEmpty ();
    }
    
    private static class Analyser {
        public BigInteger constantPart = BigInteger.ZERO;
        public ListOfMonomial parts = SLListOfMonomial.EMPTY_LIST;
        private final Services services;
        private final Operator numbers, add;
            
        public Analyser(final Services services) {
            this.services = services;
            final IntegerLDT intLDT = 
                services.getTypeConverter().getIntegerLDT();
            numbers = intLDT.getNumberSymbol();
            add     = intLDT.getArithAddition();
        }
        
        public void analyse(Term polynomial) {
            if ( polynomial.op () == add ) {
                analyse ( polynomial.sub ( 0 ) );
                analyse ( polynomial.sub ( 1 ) );
            } else if ( polynomial.op () == numbers ) {
                final BigInteger c =
                    new BigInteger ( AbstractMetaOperator
                                     .convertToDecimalString ( polynomial, services ) );
                constantPart = constantPart.add ( c );
            } else {
                parts = parts.prepend ( Monomial.create ( polynomial, services ) );
            }
        }
    }

    /**
     * @return the list of all monomials that occur in <code>a</code> but not
     *         in <code>b</code>. multiplicity is treated as well here, so
     *         this is really difference of multisets
     */
    private static ListOfMonomial difference(ListOfMonomial a, ListOfMonomial b) {
        ListOfMonomial res = a;
        final IteratorOfMonomial it = b.iterator ();
        while ( it.hasNext () && !res.isEmpty () )
            res = res.removeFirst ( it.next () );
        return res;
    }

    public BigInteger getConstantTerm() {
        return constantPart;
    }

    public ListOfMonomial getParts() {
        return parts;
    }
    
}
