package de.uka.ilkd.key.rule.updatesimplifier;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import de.uka.ilkd.key.java.Services;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.AbstractUpdateRule;
import de.uka.ilkd.key.rule.UpdateSimplifier;

/**
 * applies an update on a location dependent function. In fact this means it
 * removes all location-value pairs from the update that do not influence
 * the value of the function (incl. the function's arguments)
 */
public class ApplyOnLocationDependentFunction extends AbstractUpdateRule {

    
    /**
     * creates an instance of the update simplification rule 
     * @param updateSimplifier the UpdateSimplifier this rule belongs to
     */
    public ApplyOnLocationDependentFunction(UpdateSimplifier updateSimplifier) {
        super(updateSimplifier);
    }

    public Term apply(Update update, Term target, Services services) {
        assert target.op() instanceof LocationDependentFunction;

        final ProtectedSet protectedLocs = determineProtectedLocations(target);

        final ArrayOfAssignmentPair pairs = update.getAllAssignmentPairs();
        final LinkedList newPairs = new LinkedList();
        for (int i = 0, sz = pairs.size(); i<sz; i++) {
            final AssignmentPair pair = pairs.getAssignmentPair(i); 
            if (protectedLocs.isProtected(pair.location())) {        
                newPairs.add(pair);
            } 
        }
        
        if (pairs.size() != newPairs.size()) {
            final AssignmentPair[] ap = new AssignmentPair[newPairs.size()];
            return UpdateSimplifierTermFactory.DEFAULT.
              createUpdateTerm((AssignmentPair[])newPairs.toArray(ap), target);
        } else {
            return UpdateSimplifierTermFactory.DEFAULT.createUpdateTerm(pairs, target);
        }
    }

    /** 
     * determines all locations that may influence the 
     * value of the given term. These locations may not be deleted from the update.
     * In fact it computes a safe approximation  (a superset) of the set of protected 
     * locations.
     *    
     * @param term the Term to scan for locations that may influence its value
     * @return the set of protected values 
     */
    private ProtectedSet determineProtectedLocations(Term term) {     
        ProtectedLocationCollector collector = 
            new ProtectedLocationCollector();
        term.execPreOrder(collector); 
        return collector.getProtectedSet();
    }

    public boolean isApplicable(Update update, Term target) {
        return target.op() instanceof LocationDependentFunction;
    }
    
    
    public Term matchingCondition(Update update, Term target, Services services) {
        // should not be called for the moment
        return null;
    } 
    

    public class ProtectedLocationCollector extends Visitor {

        ProtectedSet protectedLocations = new ProtectedSet();
        
        public ProtectedLocationCollector() {}
        
        public void visit(Term visited) {
            final Operator op = visited.op();
            protectedLocations.add(op);               
            if (op instanceof LocationDependentFunction) {
                final ArrayOfLocationDescriptor descriptors = 
                    ((LocationDependentFunction)op).locationDescriptor();
                for (int i = 0, sz = descriptors.size(); i<sz; i++) {
                    final LocationDescriptor locationDescriptor = 
                        descriptors.getLocationDescriptor(i);
                    if (locationDescriptor instanceof BasicLocationDescriptor) {
                        final BasicLocationDescriptor basicLocationDescriptor = 
                            ((BasicLocationDescriptor)locationDescriptor);
                        basicLocationDescriptor.getFormula().execPreOrder(this);
                        basicLocationDescriptor.getLocTerm().execPreOrder(this);
                    } else {
                        protectedLocations.protectAll();
                    } 
                }
            }    
        }

        public ProtectedSet getProtectedSet() {          
            return protectedLocations;
        }

        
    }

    public class ProtectedSet {

        /** if set to true all locations are protected */
        private boolean protectAll = false;

        /** set of concrete locations to protect */
        private Set protectedLocations = new HashSet();
        
        /** indicates if arrays have to be protected */
        private boolean protectArrays = false;
        
        public ProtectedSet() {}
        
        public void protectAll() {
            protectAll = true;            
        }

        public void add(Operator loc) { 
            if (loc instanceof ArrayOp) {
                protectArrays = true;
            } else if (loc instanceof AttributeOp) {
                protectedLocations.add(loc);
                protectedLocations.add(((AttributeOp)loc).attribute());
            } else if (loc instanceof ProgramVariable) {
                final ProgramVariable pv = (ProgramVariable)loc;
                assert !pv.isMember() || pv.isStatic();
                protectedLocations.add(loc);
            } else if (loc instanceof NonRigidFunction && 
                    !(loc instanceof LocationDependentFunction)) {
                protectAll = true;
            } else if (loc instanceof Modality) {
                protectAll = true;
            } 
        }
        
        public boolean isProtected(Operator op) {
            if (protectAll) {
                return true;
            } else if (op instanceof ArrayOp) {
                return protectArrays;
            } else {
                return protectedLocations.contains(op);
            }
        }

        public String toString() {
            if (protectAll) return "{*}";
            String result = protectedLocations.toString();
            if (protectArrays) result += " and all arrays";
            return result;
        }
    }

}
