// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.TermBuilder;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.NonRigidFunction;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.UseMethodContractRuleFilter;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.strategy.feature.*;
import de.uka.ilkd.key.strategy.termProjection.*;
import de.uka.ilkd.key.strategy.termfeature.*;
import de.uka.ilkd.key.strategy.termgenerator.*;


/**
 * Strategy tailored to be used as long as a java program can be found in
 * the sequent.
 */
public class JavaCardDLStrategy extends AbstractFeatureStrategy {

    private static final int IN_EQ_SIMP_NON_LIN_COST = 1000;
    
    private final RuleSetDispatchFeature costComputationDispatcher;
    private final Feature costComputationF;
    private final RuleSetDispatchFeature approvalDispatcher;
    private final Feature approvalF;
    private final RuleSetDispatchFeature instantiationDispatcher;
    private final Feature instantiationF;
    
    protected final RuleSetDispatchFeature getCostComputationDispatcher() {
        return costComputationDispatcher;
    }

    protected final RuleSetDispatchFeature getInstantiationDispatcher() {
        return instantiationDispatcher;
    }

    private final StrategyProperties strategyProperties;
    
    protected JavaCardDLStrategy(Proof p_proof,
                                 StrategyProperties strategyProperties) {
        
        super ( p_proof );
        
        this.strategyProperties =
            (StrategyProperties)strategyProperties.clone ();
        
        this.tf = new ArithTermFeatures ( p_proof.getServices ()
                                          .getTypeConverter ().getIntegerLDT () );
        
        costComputationDispatcher = setupCostComputationF ( p_proof );
        approvalDispatcher = setupApprovalDispatcher ( p_proof );
        instantiationDispatcher = setupInstantiationF ( p_proof );
        
        costComputationF = setupGlobalF ( costComputationDispatcher, p_proof );
        instantiationF = setupGlobalF ( instantiationDispatcher, p_proof );
        approvalF = add ( setupApprovalF ( p_proof ), approvalDispatcher );
    }

    private Feature setupGlobalF(Feature dispatcher, Proof p_proof) {
        final Feature simplifierF = selectSimplifier ( -10000 );
        
        final Feature ifMatchedF = ifZero ( MatchedIfFeature.INSTANCE,
                                            longConst ( +1 ) );
    
//        final Feature splitF =
//            ScaleFeature.createScaled ( CountBranchFeature.INSTANCE, 400);
        final Feature ifThenElseF =
            ScaleFeature.createScaled ( IfThenElseMalusFeature.INSTANCE, 500);
    
        final Feature strengthenConstraints =
            ifHeuristics ( new String[] { "concrete", "closure" },
                           longConst ( 0 ),
                           ifZero ( ConstraintStrengthenFeatureUC.create ( p_proof ),
                                    inftyConst () ) );
    
        String methProp = strategyProperties.getProperty(
                StrategyProperties.METHOD_OPTIONS_KEY);
            
        final Feature methodSpecF;
        if (methProp.equals(StrategyProperties.METHOD_CONTRACT)) {   
            methodSpecF = methodSpecFeature(longConst(-20));
        } else if (methProp.equals(StrategyProperties.METHOD_EXPAND)) {  
            methodSpecF = methodSpecFeature(inftyConst());
        } else if (methProp.equals(StrategyProperties.METHOD_NONE)) {   
            methodSpecF = methodSpecFeature(inftyConst());
        } else throw new RuntimeException("Unexpected strategy property "+
            methProp);
        
        return SumFeature.createSum ( new Feature [] {
              AutomatedRuleFeature.INSTANCE,
//              splitF,
              dispatcher,
              NonDuplicateAppFeature.INSTANCE,
              strengthenConstraints, AgeFeature.INSTANCE,
              simplifierF, methodSpecF, ifMatchedF,
              ifThenElseF } );
    }

    private Feature methodSpecFeature(Feature cost) {
        return ConditionalFeature.createConditional(
            UseMethodContractRuleFilter.INSTANCE, cost );        
    }    

    
    
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //
    // Feature terms that handle the computation of costs for taclet applications
    //
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////

    
    private RuleSetDispatchFeature setupCostComputationF(Proof p_proof) {
        final IntegerLDT numbers =
            p_proof.getServices().getTypeConverter().getIntegerLDT();
            
        final RuleSetDispatchFeature d = RuleSetDispatchFeature.create ();
            
        bindRuleSet ( d, "closure", -9000 );
        bindRuleSet ( d, "concrete", -10000 );
        bindRuleSet ( d, "alpha", -7000 );
        bindRuleSet ( d, "delta", -6000 );
            
        bindRuleSet ( d, "pull_out_quantifier", inftyConst () );
        bindRuleSet ( d, "simplify_boolean", -200 );
        bindRuleSet ( d, "simplify", -200 );
        bindRuleSet ( d, "simplify_int", -100 );
            
        bindRuleSet ( d, "beta",
               SumFeature.createSum ( new Feature [] {
                 longConst ( -50 ),
                 ifZero ( PurePosDPathFeature.INSTANCE,
                          longConst ( -200 ) ),
                 ScaleFeature.createScaled ( CountPosDPathFeature.INSTANCE,
                                              -3.0 ),
                 ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE,
                                              10.0 ),
               } ) );
        
        bindRuleSet ( d, "test_gen", inftyConst () );

        // currently gamma rules are not applied automatically at all by this
        // strategy
        bindRuleSet ( d, "gamma", inftyConst () );
        bindRuleSet ( d, "gamma_destructive", inftyConst () );
            	
        bindRuleSet ( d, "replace_known",
           add ( ifZero ( MatchedIfFeature.INSTANCE,
                          DiffFindAndIfFeature.INSTANCE ),
                 longConst ( -4100 ),
                 ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE, 10.0 ) ) );        
            
        bindRuleSet ( d, "confluence_restricted",
                      ifZero ( MatchedIfFeature.INSTANCE,
                               DiffFindAndIfFeature.INSTANCE ) );
        
        final TermBuffer equation = new TermBuffer ();
        final TermBuffer left = new TermBuffer (), right = new TermBuffer ();

        // applying equations less deep/less leftish in terms/formulas is preferred
        // this is important for reducing polynomials (start with the biggest
        // summands)
        bindRuleSet ( d, "apply_equations",
           SumFeature.createSum ( new Feature[] {
                 ifZero ( applyTF ( FocusProjection.create ( 0 ), tf.intF ),
                          applyTF ( FocusProjection.create ( 0 ), tf.monomial ) ),
                 ifZero ( MatchedIfFeature.INSTANCE,
                   add (
                   CheckApplyEqFeature.INSTANCE,
                   let ( equation, AssumptionProjection.create ( 0 ),
                   let ( left, sub ( equation, 0 ),
                   let ( right, sub ( equation, 1 ),
                         ifZero ( applyTF ( left, tf.intF ),
                                  add ( applyTF ( left, tf.monomial ),
                                        applyTF ( right, tf.polynomial ),
                                        MonomialsSmallerThanFeature
                                        .create ( right, left, numbers ) ),
                                  TermSmallerThanFeature.create ( right, left ) )
                            ) ) ) ) ),
                 ScaleFeature.createScaled ( FindRightishFeature.INSTANCE, -5.0 ),
                 ifZero ( DirectlyBelowSymbolFeature.create ( Op.EQUALS, 0 ),
                          longConst ( -20 ) ),
                 longConst ( -4000 ) } ) );

        bindRuleSet ( d, "insert_eq_nonrigid",
                      applyTF ( FocusProjection.create ( 0 ),
                                IsNonRigidTermFeature.INSTANCE ) );
        
        bindRuleSet ( d, "order_terms",
           add ( ifZero (
                   applyTF ( "commEqLeft", tf.intF ),
                   add ( applyTF ( "commEqRight", tf.monomial ),
                         applyTF ( "commEqLeft", tf.polynomial ),
                         monSmallerThan ( "commEqLeft", "commEqRight", numbers ) ),
                   termSmallerThan ( "commEqLeft", "commEqRight" ) ),
                   longConst ( -5000 ) ) );
        
        bindRuleSet ( d, "simplify_literals",
                      ifZero ( ConstraintStrengthenFeatureUC.create(p_proof),
                               longConst ( 0 ),
                               longConst ( -8000 ) ) );
        
        bindRuleSet ( d, "simplify_instanceof_static",
                      add ( EqNonDuplicateAppFeature.INSTANCE,
                            longConst ( -500 ) ) );
        
        bindRuleSet ( d, "evaluate_instanceof", longConst ( -500 ) );
        
        bindRuleSet ( d, "beta_lemma", -10 );
        
        bindRuleSet ( d, "try_apply_subst",
                      add ( EqNonDuplicateAppFeature.INSTANCE,
                            longConst ( -10000 ) ) );
        
        bindRuleSet ( d, "split_if", 500 );
        
        final String[] exceptionsWithPenalty = new String[]{
                    "java.lang.NullPointerException",
                    "java.lang.ArrayIndexOutOfBoundsException",
                    "java.lang.ArrayStoreException",
                    "java.lang.ClassCastException"                
        };
            
        bindRuleSet ( d, "simplify_prog",
           ifZero ( ThrownExceptionFeature.create( exceptionsWithPenalty, 
                                                   p_proof.getServices() ),
                    longConst ( 500 ) ,
                    longConst ( -100 ) ) );
                
        
        // features influenced by the strategy options
        
        boolean useLoopExpand = strategyProperties.getProperty(
                StrategyProperties.LOOP_OPTIONS_KEY).
                    equals(StrategyProperties.LOOP_EXPAND);
        boolean useLoopInvariant = strategyProperties.getProperty(
                StrategyProperties.LOOP_OPTIONS_KEY).
                    equals(StrategyProperties.LOOP_INVARIANT);
        boolean useQueryExpand = strategyProperties.getProperty(
                StrategyProperties.QUERY_OPTIONS_KEY).
                    equals(StrategyProperties.QUERY_EXPAND); 
        boolean programsToRight = useQueryExpand ||
                strategyProperties.getProperty(
                StrategyProperties.QUERY_OPTIONS_KEY).
                    equals(StrategyProperties.QUERY_PROGRAMS_TO_RIGHT);
        
        String methProp =
            strategyProperties.getProperty ( StrategyProperties.METHOD_OPTIONS_KEY );
	
        if (methProp.equals(StrategyProperties.METHOD_CONTRACT)) {
            bindRuleSet(d, "method_expand", longConst(100));	
        } else if (methProp.equals(StrategyProperties.METHOD_EXPAND)) {
            bindRuleSet(d, "method_expand", longConst(100));	   
        } else if (methProp.equals(StrategyProperties.METHOD_NONE)) {
            bindRuleSet(d, "method_expand", inftyConst());	  
        } else throw new RuntimeException("Unexpected strategy property "+
                                          methProp);
	
        
        bindRuleSet ( d, "loop_expand",
                      useLoopExpand ? longConst ( 0 )
                                    : inftyConst () );
        
        bindRuleSet  ( d, "loop_invariant", 
                       useLoopInvariant ? longConst ( 100 )  
                                        : inftyConst () );
            
        bindRuleSet  ( d, "loop_invariant_proposal", 
                       ifHeuristics(new String[]{"loop_invariant"}, 
                                    longConst(0), inftyConst()));
            
        
        // query treatment
	// attention: usually this application is against the term order but
	// does not interfere as only applied below updates
        bindRuleSet ( d, "query_normalize", 
           ifZero ( ifZero(DepthGreaterThanFeature.create(1),
			   add ( DirectlyBelowOpClassFeature.create
				 ( NonRigidFunction.class ),
				 applyTF (FocusProjection.create(2),
					  IsUpdateOperatorTF.INSTANCE),
				 applyTF("t", IsNonRigidTermFeature.INSTANCE)),
			   inftyConst()),
                    longConst (-5000),
                    inftyConst () ) );

        // delete cast
        bindRuleSet ( d, "cast_deletion",
                      ifZero ( implicitCastNecessary ( instOf ( "castedTerm" ) ),
                               longConst ( -5000 ), inftyConst () ) );
                
        bindRuleSet ( d, "queries", 
           useQueryExpand ? longConst(5000) : inftyConst() );

        bindRuleSet ( d, "inReachableStateImplication",
                      add ( NonDuplicateAppModPositionFeature.INSTANCE,
                            longConst ( 100 ) ) );
        
        if ( programsToRight )
            bindRuleSet ( d, "boxDiamondConv",
                          add ( TopLevelFindFeature.ANTEC_WITH_UPDATE,
                                longConst ( 10000 ) ) );
        else
            bindRuleSet ( d, "boxDiamondConv", inftyConst () );
        
        bindRuleSet ( d, "cut", not ( isInstantiated ( "cutFormula" ) ) );
        
        setupArithPrimaryCategories ( d );
        setupPolySimp ( d, numbers );        
        setupInEqSimp ( d, numbers );
            
        setupSystemInvariantSimp(d);
        
        // For taclets that need instantiation, but where the instantiation is
        // deterministic and does not have to be repeated at a later point, we
        // setup the same feature terms as in the instantiation method. The
        // definitions in <code>setupInstantiationWithoutRetry</code> should
        // give cost infinity to those incomplete rule applications that will
        // never be instantiated (so that these applications can be removed from
        // the queue and do not have to be considered again).
        setupInstantiationWithoutRetry ( d, p_proof );
        
        return d;
    }

    private void setupArithPrimaryCategories(RuleSetDispatchFeature d) {
        // Buchberger's algorithmus for handling polynomial equations over
        // the integers
            
        bindRuleSet ( d, "polySimp_expand", -4500 );
        bindRuleSet ( d, "polySimp_directEquations", -3000 );
        bindRuleSet ( d, "polySimp_pullOutGcd", -2250 );
        bindRuleSet ( d, "polySimp_saturate", -2000 );

        // Omega test for handling linear arithmetic and inequations over the
        // integers
            
        bindRuleSet ( d, "inEqSimp_expand", -4500 );
        bindRuleSet ( d, "inEqSimp_directInEquations", -3000 );
        bindRuleSet ( d, "inEqSimp_propagation", -2500 );
        bindRuleSet ( d, "inEqSimp_pullOutGcd", -2250 );
        bindRuleSet ( d, "inEqSimp_saturate", -2000 );
        bindRuleSet ( d, "inEqSimp_special_nonLin", -1400 );
        
        if ( arithNonLinInferences () )
            bindRuleSet ( d, "inEqSimp_nonLin", IN_EQ_SIMP_NON_LIN_COST );
        else
            bindRuleSet ( d, "inEqSimp_nonLin", inftyConst () );
    }

    private boolean arithNonLinInferences() {
        return !StrategyProperties.NON_LIN_ARITH_NONE.equals (
                 strategyProperties.getProperty
                 ( StrategyProperties.NON_LIN_ARITH_OPTIONS_KEY ) );
    }

    private void setupPolySimp( RuleSetDispatchFeature d, IntegerLDT numbers) {
        
        // category "expansion" (normalising polynomial terms)
        
        bindRuleSet ( d, "polySimp_elimSubNeg", longConst ( -120 ) );

        bindRuleSet ( d, "polySimp_homo",
          add ( applyTF ( "homoRight",
                          add ( not ( tf.zeroLiteral ), tf.polynomial ) ),
                or ( applyTF ( "homoLeft", or ( tf.addF, tf.negMonomial ) ),
                     not ( monSmallerThan ( "homoRight", "homoLeft", numbers) ) ),
                longConst ( -120 ) ) );
        
        bindRuleSet ( d, "polySimp_pullOutFactor",
                      add ( applyTFNonStrict ( "pullOutLeft", tf.literal ),
                            applyTFNonStrict ( "pullOutRight", tf.literal ),
                            longConst ( -120 ) ) );

        bindRuleSet ( d, "polySimp_elimOneLeft", -120 );

        bindRuleSet ( d, "polySimp_elimOneRight", -120 );

        bindRuleSet ( d, "polySimp_mulOrder",
           add ( applyTF ( "commRight", tf.monomial ),
                 or ( applyTF ( "commLeft", tf.addF ),
       	              add ( applyTF ( "commLeft", tf.atom ),
                            atomSmallerThan ( "commLeft", "commRight", numbers ) ) ),
                 longConst ( -100 ) ) );

        bindRuleSet ( d, "polySimp_mulAssoc",
                      SumFeature.createSum( new Feature[] {
                            applyTF ( "mulAssocMono0", tf.monomial ),
                            applyTF ( "mulAssocMono1", tf.monomial ),
                            applyTF ( "mulAssocAtom", tf.atom ),
                            longConst ( -80 ) } ) );

        bindRuleSet ( d, "polySimp_addOrder",
           SumFeature.createSum( new Feature[] {
             applyTF ( "commLeft", tf.monomial ),
             applyTF ( "commRight", tf.polynomial ),
             monSmallerThan ( "commRight", "commLeft", numbers),
             longConst ( -60 ) } ) );

        
        bindRuleSet ( d, "polySimp_addAssoc",
                      SumFeature.createSum( new Feature[] {
                            applyTF ( "addAssocPoly0", tf.polynomial ),
                            applyTF ( "addAssocPoly1", tf.polynomial ),
                            applyTF ( "addAssocMono", tf.monomial ),
                            longConst ( -10 ) } ) );

        bindRuleSet ( d, "polySimp_dist",
                      SumFeature.createSum( new Feature[] {
                        longConst ( -35 ),
                        ifZero ( applyTF ( "distCoeff", tf.monomial ),
                                 longConst ( -15 ),
                                 applyTF ( "distCoeff", tf.polynomial ) ),
                        applyTF ( "distSummand0", tf.polynomial ),
                        applyTF ( "distSummand1", tf.polynomial ) } ) );

        bindRuleSet ( d, "polySimp_divDistr",
           SumFeature.createSum ( new Feature[] {
             applyTF ( "divDistrDenom", tf.literal ),
             applyTFNonStrict ( "divDistrNum", tf.literal ),
             applyTFNonStrict ( "divDistrRem", tf.polynomial ),
             applyTFNonStrict ( "divDistrCoeff", tf.nonCoeffMonomial ),
             ifZero ( isInstantiated ( "divDistrNum" ),
                      PolynomialValuesCmpFeature
                      .divides ( instOf ( "divDistrNum" ),
                                 instOf ( "divDistrDenom" ) ) ) } ) );                   

        // category "direct equations"
        
        bindRuleSet ( d, "polySimp_balance",
           SumFeature.createSum ( new Feature[] {
             applyTF ( "sepResidue", tf.polynomial ),
             ifZero ( isInstantiated ( "sepPosMono" ),
               add ( applyTF ( "sepPosMono", tf.nonNegMonomial ),
                     monSmallerThan ( "sepResidue", "sepPosMono", numbers ) ) ),
             ifZero ( isInstantiated ( "sepNegMono" ),
               add ( applyTF ( "sepNegMono", tf.negMonomial ),
                     monSmallerThan ( "sepResidue", "sepNegMono", numbers ) ) ),
             longConst ( -30 )
           } ) );
        
        bindRuleSet ( d, "polySimp_normalise",
                      add ( applyTF ( "invertRight", tf.zeroLiteral ),
                            applyTF ( "invertLeft", tf.negMonomial ),
                            longConst ( -30 ) ) );
        
        // application of equations: some specialised rules that handle
        // monomials and their coefficients properly
        
        bindRuleSet ( d, "polySimp_applyEq",
           add ( not ( DirectlyBelowSymbolFeature.create ( tf.mul ) ),
                 ifZero ( MatchedIfFeature.INSTANCE,
                          ReducibleMonomialsFeature
                          	.create ( instOf ( "applyEqDividend" ),
                          	          instOf ( "applyEqDivisor" ) ) ),
                 longConst ( 1 ) ) );

        bindRuleSet ( d, "polySimp_applyEqRigid",
           add ( not ( DirectlyBelowSymbolFeature.create ( tf.mul ) ),
                 ifZero ( MatchedIfFeature.INSTANCE,
                          ReducibleMonomialsFeature
                          	.create ( instOf ( "applyEqDividend" ),
                          	          instOf ( "applyEqDivisorr" ) ) ),
                 longConst ( 2 ) ) );

        // category "saturate"
        
        bindRuleSet ( d, "polySimp_critPair",
           ifZero ( MatchedIfFeature.INSTANCE,
                    add ( monSmallerThan ( "cpLeft1", "cpLeft2", numbers ),
                          not ( TrivialMonomialLCRFeature
                                .create ( instOf ( "cpLeft1" ),
                                          instOf ( "cpLeft2" ) ) ) ) ) );
                   
    }
    

    // For taclets that need instantiation, but where the instantiation is
    // deterministic and does not have to be repeated at a later point, we
    // setup the same feature terms as in the instantiation method. The
    // definitions in <code>setupInstantiationWithoutRetry</code> should
    // give cost infinity to those incomplete rule applications that will
    // never be instantiated (so that these applications can be removed from
    // the queue and do not have to be considered again).
    private void setupPolySimpInstantiationWithoutRetry(RuleSetDispatchFeature d,
                                                        Proof p_proof) {
        // category "direct equations"

        setupPullOutGcd ( d, "polySimp_pullOutGcd", p_proof );
    }

    private Feature termSmallerThan(String smaller, String bigger) {
        return
        TermSmallerThanFeature.create ( instOf ( smaller ), instOf ( bigger ) );
    }

    private Feature monSmallerThan(String smaller, String bigger, IntegerLDT numbers) {
        return
        MonomialsSmallerThanFeature.create ( instOf ( smaller ), instOf ( bigger ),
                                             numbers );
    }

    private Feature atomSmallerThan(String smaller, String bigger, IntegerLDT numbers) {
        return
        AtomsSmallerThanFeature.create ( instOf ( smaller ), instOf ( bigger ),
                                         numbers );
    }

    
    private void setupPullOutGcd(RuleSetDispatchFeature d,
                                 String ruleSet,
                                 Proof p_proof) {
        final TermBuffer gcd = new TermBuffer ();
        final TermBuffer two = new TermBuffer ();
        two.setContent ( TermBuilder.DF.zTerm ( p_proof.getServices (), "2" ) );
        
        bindRuleSet ( d, ruleSet,
           add ( applyTF ( "elimGcdLeft", tf.nonNegMonomial ),
                 applyTF ( "elimGcdRight", tf.polynomial ),
                 let ( gcd,
                       CoeffGcdProjection.create ( instOf ( "elimGcdLeft" ),
                                                   instOf ( "elimGcdRight" ) ),
                       add ( PolynomialValuesCmpFeature.leq ( two, gcd ),
                             instantiate ( "elimGcd", gcd ) ) ) ) );
    }    

    
    private void setupInEqSimp(RuleSetDispatchFeature d, IntegerLDT numbers) {
        
        // category "expansion" (normalising inequations)
 
        bindRuleSet ( d, "inEqSimp_moveLeft", -80 );

        bindRuleSet ( d, "inEqSimp_commute",
                      SumFeature.createSum ( new Feature[] {
                        applyTF ( "commRight", tf.monomial ),
                        applyTF ( "commLeft", tf.polynomial ),
                        monSmallerThan ( "commLeft", "commRight", numbers ),
                        longConst ( -40 ) } ) );

        // this is copied from "polySimp_homo"
        bindRuleSet ( d, "inEqSimp_homo",
           add ( applyTF ( "homoRight",
                             add ( not ( tf.zeroLiteral ), tf.polynomial ) ),
                 or ( applyTF ( "homoLeft", or ( tf.addF, tf.negMonomial ) ),
                      not ( monSmallerThan ( "homoRight", "homoLeft", numbers) ) ) ) );

        // category "direct inequations"

        // this is copied from "polySimp_balance"
        bindRuleSet ( d, "inEqSimp_balance",
           add (
             applyTF ( "sepResidue", tf.polynomial ),
             ifZero ( isInstantiated ( "sepPosMono" ),
                add ( applyTF ( "sepPosMono", tf.nonNegMonomial ),
                      monSmallerThan ( "sepResidue", "sepPosMono", numbers ) ) ),
             ifZero ( isInstantiated ( "sepNegMono" ),
                add ( applyTF ( "sepNegMono", tf.negMonomial ),
                      monSmallerThan ( "sepResidue", "sepNegMono", numbers ) ) )
             ) );
        
        // this is copied from "polySimp_normalise"
        bindRuleSet ( d, "inEqSimp_normalise",
                      add ( applyTF ( "invertRight", tf.zeroLiteral ),
                            applyTF ( "invertLeft", tf.negMonomial ) ) );

        // category "saturate"
        
        bindRuleSet ( d, "inEqSimp_antiSymm", longConst ( -20 ) );

        bindRuleSet ( d, "inEqSimp_exactShadow",
           ifZero ( MatchedIfFeature.INSTANCE,
                    SumFeature.createSum ( new Feature[] {
                      applyTF ( "esLeft", tf.nonCoeffMonomial ),
                      applyTFNonStrict ( "esCoeff1", tf.nonNegLiteral ),
                      applyTFNonStrict ( "esCoeff2", tf.nonNegLiteral ),
                      applyTF ( "esRight1", tf.polynomial ),
                      applyTF ( "esRight2", tf.polynomial ),
                      not ( PolynomialValuesCmpFeature
                            .leq ( instOf ( "esRight2" ),
                                   instOf ( "esRight1" ),
                                   instOfNonStrict ( "esCoeff1" ),
                                   instOfNonStrict ( "esCoeff2" ) ))
                    } ) ));              

        // category "propagation"

        bindRuleSet ( d, "inEqSimp_contradInEqs",
           ifZero ( MatchedIfFeature.INSTANCE,
             SumFeature.createSum ( new Feature[] {
               DiffFindAndIfFeature.INSTANCE,
               applyTFNonStrict ( "contradCoeffSmaller", tf.posLiteral ),
               applyTFNonStrict ( "contradCoeffBigger", tf.posLiteral ),
               PolynomialValuesCmpFeature
               .lt ( instOf ( "contradRightSmaller" ),
                     instOf ( "contradRightBigger" ),
                     instOfNonStrict ( "contradCoeffBigger" ),
                     instOfNonStrict ( "contradCoeffSmaller" ) ) } ) ) );

        bindRuleSet ( d, "inEqSimp_contradEqs",
           add ( ifZero ( MatchedIfFeature.INSTANCE,
                          PolynomialValuesCmpFeature
                          .lt ( instOf ( "contradRightSmaller" ),
                                instOf ( "contradRightBigger" ) ) ),
                 longConst ( -60 ) ) );

        bindRuleSet ( d, "inEqSimp_strengthen", longConst ( -30 ) );

        bindRuleSet ( d, "inEqSimp_subsumption",
           ifZero ( MatchedIfFeature.INSTANCE,
             SumFeature.createSum ( new Feature[] {
               DiffFindAndIfFeature.INSTANCE,
               applyTFNonStrict ( "subsumCoeffSmaller", tf.posLiteral ),
               applyTFNonStrict ( "subsumCoeffBigger", tf.posLiteral ),
               PolynomialValuesCmpFeature
               .leq ( instOf ( "subsumRightSmaller" ),
                      instOf ( "subsumRightBigger" ),
                      instOfNonStrict ( "subsumCoeffBigger" ),
                      instOfNonStrict ( "subsumCoeffSmaller" ) ) } ) ) );

        // category "handling of non-linear inequations"
        
        if ( arithNonLinInferences () ) {
            setupMultiplyInequations ( d, longConst ( 1000 ), numbers );
         
            bindRuleSet ( d, "inEqSimp_split_eq",
                          add ( TopLevelFindFeature.SUCC, longConst ( -100 ) ) );
        }
    }

    private void setupMultiplyInequations(RuleSetDispatchFeature d,
                                          Feature notAllowedF,
                                          IntegerLDT numbers) {
        final TermBuffer intRel = new TermBuffer ();
        
        final Feature unboundedMult =
            sum ( intRel, SequentFormulasGenerator.sequent (),
                  not ( add ( applyTF ( intRel, tf.intRelation ),
                              AllowInEquationMultiplication
                              .create ( instOf ( "multLeft" ),
                                        instOf ( "multFacLeft" ),
                                        sub ( intRel, 0 ) ) ) ) );
        
        final Feature boundedMult =
            not ( sum ( intRel, SequentFormulasGenerator.sequent (),
                        not ( add ( applyTF ( intRel, tf.intRelation ),
                                    BoundedInEquationsFeature
                                    .create ( instOf ( "multLeft" ),
                                              instOf ( "multFacLeft" ),
                                              sub ( intRel, 0 ) ) ) ) ) );
        
        bindRuleSet ( d, "inEqSimp_nonLin_multiply",
           ifZero ( MatchedIfFeature.INSTANCE,
              SumFeature.createSum ( new Feature[] {
                applyTF ( "multLeft", tf.nonNegMonomial ),
                applyTF ( "multFacLeft", tf.nonNegMonomial ),
                ifZero ( applyTF ( "multRight", tf.literal ),
                         longConst ( -100 ),
                         applyTF ( "multRight", tf.polynomial ) ),
                ifZero ( applyTF ( "multFacRight", tf.literal ),
                         longConst ( -100 ),
                         applyTF ( "multFacRight", tf.polynomial ) ),
                not ( TermSmallerThanFeature.create
                      ( FocusProjection.create ( 0 ),
                        AssumptionProjection.create ( 0 ) ) ),
                ifZero ( unboundedMult, notAllowedF ),
                applyTF ( "multLeft", rec ( tf.mulF, longTermConst ( 20 ) ) ),
                applyTF ( "multFacLeft", rec ( tf.mulF, longTermConst ( 20 ) ) ),
                ifZero ( not ( boundedMult ), longConst ( 400 ) )
              } ) ) );
    }
    
    private void setupInEqSimpInstantiation(RuleSetDispatchFeature d,
                                            Proof p_proof) {
        // category "handling of non-linear inequations"

        setupSquaresAreNonNegative ( d );
        
        if ( arithNonLinInferences() )            
            setupInEqCaseDistinctions ( d, p_proof );
    }

    // For taclets that need instantiation, but where the instantiation is
    // deterministic and does not have to be repeated at a later point, we
    // setup the same feature terms as in the instantiation method. The
    // definitions in <code>setupInstantiationWithoutRetry</code> should
    // give cost infinity to those incomplete rule applications that will
    // never be instantiated (so that these applications can be removed from
    // the queue and do not have to be considered again).
    private void setupInEqSimpInstantiationWithoutRetry(RuleSetDispatchFeature d,
                                                        Proof p_proof) {
        // category "direct inequations"

        setupPullOutGcd ( d, "inEqSimp_pullOutGcd", p_proof );
        
        // category "handling of non-linear inequations"
        
        final TermBuffer divisor = new TermBuffer ();
        final TermBuffer dividend = new TermBuffer ();
            
        bindRuleSet ( d, "inEqSimp_nonLin_divide",
          SumFeature.createSum ( new Feature[] {
            applyTF ( "divProd", tf.nonCoeffMonomial ),
            applyTFNonStrict ( "divProdBoundNonPos", tf.nonPosLiteral ),
            applyTFNonStrict ( "divProdBoundNonNeg", tf.nonNegLiteral ),
            ifZero ( MatchedIfFeature.INSTANCE,
              let ( divisor, instOf ( "divX" ),
              let ( dividend, instOf ( "divProd" ),
              SumFeature.createSum ( new Feature[] {
                applyTF ( divisor, tf.nonCoeffMonomial ),
                not ( applyTF ( dividend, eq ( divisor ) ) ),
                applyTFNonStrict ( "divXBoundPos", tf.posLiteral ),
                applyTFNonStrict ( "divXBoundNeg", tf.negLiteral ),
                ReducibleMonomialsFeature.create ( dividend, divisor ),
                instantiate ( "divY", DivideMonomialsProjection
                                      .create ( dividend, divisor ) )
              } ) ) ) )
          } ) );

        setupNonLinTermIsPosNeg ( d, "inEqSimp_nonLin_pos", true );
        setupNonLinTermIsPosNeg ( d, "inEqSimp_nonLin_neg", false );
    }

    private void setupNonLinTermIsPosNeg(RuleSetDispatchFeature d,
                                         String ruleSet,
                                         boolean pos) {
        final TermBuffer divisor = new TermBuffer ();
        final TermBuffer dividend = new TermBuffer ();
        final TermBuffer quotient = new TermBuffer ();
        final TermBuffer antecFor = new TermBuffer ();

        bindRuleSet ( d, ruleSet,
          SumFeature.createSum ( new Feature[] {
            applyTF ( "divProd", tf.nonCoeffMonomial ),
            applyTFNonStrict ( "divProdBoundPos", tf.posLiteral ),
            applyTFNonStrict ( "divProdBoundNeg", tf.negLiteral ),
            ifZero ( MatchedIfFeature.INSTANCE,
              let ( divisor, instOf ( "divX" ),
              let ( dividend, instOf ( "divProd" ),
              SumFeature.createSum ( new Feature[] {
                applyTF ( divisor, tf.nonCoeffMonomial ),
                not ( applyTF ( dividend, eq ( divisor ) ) ),
                applyTFNonStrict ( "divXBoundNonPos", tf.nonPosLiteral ),
                applyTFNonStrict ( "divXBoundNonNeg", tf.nonNegLiteral ),
                ReducibleMonomialsFeature.create ( dividend, divisor ),
                let ( quotient,
                      DivideMonomialsProjection.create ( dividend, divisor ),
                      add ( sum ( antecFor,
                                  SequentFormulasGenerator.antecedent (),
                                  not ( applyTF ( antecFor,
                                                  pos ?
                                                  opSub ( tf.geq,
                                                          eq ( quotient ),
                                                          tf.posLiteral )
                                                  :
                                                  opSub ( tf.leq,
                                                          eq ( quotient ),
                                                          tf.negLiteral ) ) ) ),
                            instantiate ( "divY", quotient ) ) )
              } ) ) ) )
          } ) );
    }

    private void setupSquaresAreNonNegative(RuleSetDispatchFeature d) {
        final TermBuffer intRel  = new TermBuffer ();
        final TermBuffer product = new TermBuffer ();
        final TermBuffer factor  = new TermBuffer ();

        final Feature productContainsSquare =
            applyTF ( sub ( product, 0 ),
                      or ( eq ( factor ),
                           opSub ( tf.mul, any (), eq ( factor ) ) ) );
        final Feature productIsProduct =
            applyTF ( product, opSub ( tf.mul, any (), not ( tf.mulF ) ) );
        
        bindRuleSet ( d, "inEqSimp_nonNegSquares",
          forEach ( intRel, SequentFormulasGenerator.sequent (),
             ifZero ( applyTF ( intRel, tf.intRelation ),
                forEach ( product,
                          SubtermGenerator.create ( sub ( intRel, 0 ), tf.mulF ),
                   ifZero ( productIsProduct,
                            let ( factor, sub ( product, 1 ),
                                  ifZero ( productContainsSquare,
                                           instantiate ( "squareFac", factor ) ) )
                            ) ) ) ) );
    }


    private void setupInEqCaseDistinctions(RuleSetDispatchFeature d,
                                           Proof p_proof) {
        final TermBuffer intRel = new TermBuffer ();
        final TermBuffer atom = new TermBuffer ();
        final TermBuffer zero = new TermBuffer ();
        zero.setContent ( TermBuilder.DF.zTerm ( p_proof.getServices (), "0" ) );
        final TermBuffer rootInf = new TermBuffer ();

        final Feature posNegSplitting =
           forEach ( intRel, SequentFormulasGenerator.antecedent (),
             add ( applyTF ( intRel, tf.intRelation ),
                   forEach ( atom,
                             SubtermGenerator.create ( sub ( intRel, 0 ), tf.mulF ),
                      SumFeature.createSum ( new Feature[] {
                        applyTF ( atom, add ( tf.atom, not ( tf.literal ) ) ),
                        allowPosNegCaseDistinction ( atom ),
                        instantiate ( "cutFormula",
                                      opTerm ( tf.geq, atom, zero ) ),
                        longConst ( IN_EQ_SIMP_NON_LIN_COST + 300 ),
                        applyTF ( atom, rec ( any (), longTermConst ( 5 ) ) )
                      } ) ) ) );
        
        final Feature strengthening =
            forEach ( intRel, SequentFormulasGenerator.antecedent (),
              SumFeature.createSum ( new Feature[] {
                    applyTF ( intRel, add ( or ( tf.geqF, tf.leqF ),
                                            sub ( tf.atom, tf.literal ) ) ),
                    instantiate ( "cutFormula",
                                  opTerm ( Op.EQUALS,
                                           sub ( intRel, 0 ), sub ( intRel, 1 ) ) ),
                    longConst ( IN_EQ_SIMP_NON_LIN_COST + 500 ),
                    applyTF ( atom, rec ( any (), longTermConst ( 5 ) ) )
              } ) );
        
        final Feature rootInferences =
            forEach ( intRel, SequentFormulasGenerator.antecedent (),
              add ( isRootInferenceProducer ( intRel ),
                    forEach ( rootInf, RootsGenerator.create ( intRel ),
                              add ( instantiate ( "cutFormula", rootInf ),
                                    ifZero ( applyTF ( rootInf, op ( Op.OR ) ),
                                             longConst ( 50 ) ),
                                    ifZero ( applyTF ( rootInf, op ( Op.AND ) ),
                                             longConst ( 20 ) ) ) ),
                    longConst ( IN_EQ_SIMP_NON_LIN_COST )
              ) );
        
        bindRuleSet ( d, "cut", oneOf ( new Feature[] { posNegSplitting,
                                                        strengthening,
                                                        rootInferences } ) ); 
    }

    private Feature isRootInferenceProducer(TermBuffer intRel) {
        return applyTF ( intRel, add ( tf.intRelation,
                                        sub ( tf.nonCoeffMonomial,
                                              tf.literal ) ) );
    }

    private Feature allowPosNegCaseDistinction(TermBuffer atom) {
        final TermBuffer antecFor = new TermBuffer ();

        return
         sum ( antecFor, SequentFormulasGenerator.antecedent (),
               not ( applyTF ( antecFor,
                               add ( tf.intRelation,
                                     sub ( eq ( atom ), any () ),
                                     or ( op ( Op.EQUALS ),
                                          sub ( any (), tf.literal ) ) ) ) ) );
    }

    private Feature allowInEqStrengthening(TermBuffer atom, TermBuffer literal) {
        final TermBuffer antecFor  = new TermBuffer ();

        return
         not ( sum ( antecFor, SequentFormulasGenerator.antecedent (),
                     not ( applyTF ( antecFor,
                                     add ( or ( tf.leqF, tf.geqF ),
                                           sub ( eq ( atom ), eq ( literal ) ) ) ) ) ) );
    }

    private void setupInEqCaseDistinctionsApproval(RuleSetDispatchFeature d) {
        final TermBuffer atom = new TermBuffer ();
        final TermBuffer literal = new TermBuffer ();
        final TermBuffer intRel = new TermBuffer ();
        final TermBuffer rootInf = new TermBuffer ();
        
        // this is somewhat ugly. we should introduce some concept of "tagging"
        // rule application so that they can be recognised again later
        bindRuleSet ( d, "cut",
           add ( isInstantiated ( "cutFormula" ),
            or ( not ( sum ( intRel, SequentFormulasGenerator.antecedent (),
                             ifZero ( isRootInferenceProducer ( intRel ),
                                      sum ( rootInf,
                                            RootsGenerator.create ( intRel ),
                                            not ( eq ( instOf ( "cutFormula" ),
                                                       rootInf ) ) ) ) ) ),
                 add (
                 ifZero ( applyTF ( "cutFormula",
                                    opSub ( tf.geq, tf.atom, tf.zeroLiteral ) ),
                          let ( atom, sub ( instOf ( "cutFormula" ), 0 ),
                                allowPosNegCaseDistinction ( atom ) ) ),
                 ifZero ( applyTF ( "cutFormula",
                                    opSub ( Op.EQUALS, tf.atom, tf.literal ) ),
                          let ( atom, sub ( instOf ( "cutFormula" ), 0 ),
                          let ( literal, sub ( instOf ( "cutFormula" ), 1 ),
                                allowInEqStrengthening ( atom, literal ) ) ) ) ) ) ) );
    }


    
    private void setupSystemInvariantSimp(RuleSetDispatchFeature d) {
        bindRuleSet ( d, "system_invariant", 
           ifZero ( MatchedIfFeature.INSTANCE,
                    add ( applyTF ( "negLit", tf.negLiteral ),
                          applyTFNonStrict ( "nonNegLit", tf.nonNegLiteral ) ) ) );
    }


    
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //
    // Feature terms that handle the approval of complete taclet applications
    //
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////

    private Feature setupApprovalF(Proof p_proof) {
        return add ( NonDuplicateAppFeature.INSTANCE,
                     UCIncompatibleFeature.create ( p_proof ) );
    }
    
    private RuleSetDispatchFeature setupApprovalDispatcher(Proof p_proof) {
        final IntegerLDT numbers =
            p_proof.getServices().getTypeConverter().getIntegerLDT();

        final RuleSetDispatchFeature d = RuleSetDispatchFeature.create ();

        if ( arithNonLinInferences () )
            setupMultiplyInequations ( d, inftyConst (), numbers );

        // these taclets are not supposed to be applied with "squareFac" or
        // "divY" being instantiated with a metavariable
        bindRuleSet ( d, "inEqSimp_nonNegSquares", isInstantiated ( "squareFac" ) );
        bindRuleSet ( d, "inEqSimp_nonLin_divide", isInstantiated ( "divY" ) );
        bindRuleSet ( d, "inEqSimp_nonLin_pos", isInstantiated ( "divY" ) );
        bindRuleSet ( d, "inEqSimp_nonLin_neg", isInstantiated ( "divY" ) );
        
        if ( arithNonLinInferences () )
            setupInEqCaseDistinctionsApproval ( d );
        
        bindRuleSet ( d, "inReachableStateImplication",
                      NonDuplicateAppModPositionFeature.INSTANCE );

        return d;
    }

    
    
    
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////
    //
    // Feature terms that handle the instantiation of incomplete taclet
    // applications
    //
    ////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////


    private RuleSetDispatchFeature setupInstantiationF(Proof p_proof) {
        enableInstantiate ();
        
        final RuleSetDispatchFeature d = RuleSetDispatchFeature.create ();
        
//      Just an example ... as long as there are no real applications of these
//      features
//
//        final TermBuffer x = new TermBuffer (), y = new TermBuffer ();
//  
//        bindRuleSet ( d, "cut_test",
//                      forEach ( x, new Enumerate1ToNEq ( 2 ),
//                      forEach ( y, new Enumerate1ToNEq ( 3 ),
//                                add ( instantiate ( "b", x ),
//                                      instantiate ( "c", y ),
//                                      longConst ( 500 ) ) ) ) );

        setupArithPrimaryCategories ( d );
        
        setupInstantiationWithoutRetry ( d, p_proof );

        setupInEqSimpInstantiation ( d, p_proof );
        
        disableInstantiate ();
        return d;
    }

    /**
     * For taclets that need instantiation, but where the instantiation is
     * deterministic and does not have to be repeated at a later point, we setup
     * the same feature terms both in the cost computation method and in the
     * instantiation method. The definitions in
     * <code>setupInstantiationWithoutRetry</code> should give cost infinity
     * to those incomplete rule applications that will never be instantiated (so
     * that these applications can be removed from the queue and do not have to
     * be considered again).
     */
    private void setupInstantiationWithoutRetry(RuleSetDispatchFeature d,
                                                Proof p_proof) {
        setupPolySimpInstantiationWithoutRetry ( d, p_proof );
        setupInEqSimpInstantiationWithoutRetry ( d, p_proof );
    }

    
    public Name name () {
        return new Name("JavaCardDLStrategy");
    }


    /**
	 * Evaluate the cost of a <code>RuleApp</code>.
	 * 
	 * @return the cost of the rule application expressed as a
	 *         <code>RuleAppCost</code> object.
	 *         <code>TopRuleAppCost.INSTANCE</code> indicates that the rule
	 *         shall not be applied at all (it is discarded by the strategy).
	 */
    public final RuleAppCost computeCost (RuleApp app,
                                          PosInOccurrence pio,
                                          Goal goal) {
        return costComputationF.compute ( app, pio, goal );
    }

    /**
	 * Re-Evaluate a <code>RuleApp</code>. This method is called immediately
	 * before a rule is really applied
	 * 
	 * @return true iff the rule should be applied, false otherwise
	 */
    public final boolean isApprovedApp (RuleApp app,
                                        PosInOccurrence pio,
                                        Goal goal) {
        return !( approvalF.compute ( app, pio, goal ) instanceof TopRuleAppCost );
    }
    
    protected final RuleAppCost instantiateApp(RuleApp app,
                                               PosInOccurrence pio,
                                               Goal goal) {
        return instantiationF.compute ( app, pio, goal );
    }
    

    public static class Factory extends StrategyFactory {
        public Factory () {
        }

        public Strategy create ( Proof p_proof, 
                StrategyProperties strategyProperties) {     
            return new JavaCardDLStrategy ( p_proof, strategyProperties );
        }
        
        public Name name () {
            return new Name("JavaCardDLStrategy");
        }
    }

    private final ArithTermFeatures tf;
    
    private class ArithTermFeatures {

        public ArithTermFeatures (IntegerLDT numbers){
            Z = numbers.getNumberSymbol ();

            add = numbers.getArithAddition();
            mul = numbers.getArithMultiplication ();
            
            leq = numbers.getLessOrEquals ();
            geq = numbers.getGreaterOrEquals ();
            
            intS = numbers.getNumberSymbol ().sort ();
            
            intF = extendsTrans ( intS );
            
            addF = op(add);
            mulF = op(mul);
            
            geqF = op(geq);
            leqF = op(leq);

            literal =
                op ( Z );
            negLiteral =
                opSub ( Z, op ( numbers.getNegativeNumberSign() ) );
            nonNegLiteral =
                opSub ( Z, not ( op ( numbers.getNegativeNumberSign() ) ) );
            zeroLiteral =
                opSub ( Z, opSub ( numbers.getNumberLiteralFor(0), 
                                   op ( numbers.getNumberTerminator () ) ) );
            nonPosLiteral = or ( zeroLiteral, negLiteral );
            posLiteral = add ( nonNegLiteral, not ( zeroLiteral ) );

            atom = add ( not ( addF ), not ( mulF ) );
            
            // left-associatively arranged monomials, literals are only allowed
            // as right-most term
            monomial =
                or ( atom,
                     opSub ( mul,
                             rec ( mulF, or ( opSub ( mul, any (), not ( mulF ) ),
                                              add ( not ( addF ), not ( literal ) ) ) ),
                             atom ) );
            // left-associatively arranged polynomials
            polynomial = rec ( addF, or ( opSub ( add, any (), not ( addF ) ),
                                          monomial ) );
                  
            nonNegMonomial = add ( monomial,
                                   or ( not ( mulF ),
                                        sub ( any (), not ( negLiteral ) ) ) );
            negMonomial = opSub ( mul, monomial, negLiteral );            
            nonCoeffMonomial = add ( monomial,
                                     or ( not ( mulF ),
                                          sub ( any (), not ( literal ) ) ) );
            
            intEquation = opSub ( Op.EQUALS, nonNegMonomial, polynomial );
            intInEquation = add ( or ( op ( leq ), op ( geq ) ),
                                  sub ( nonNegMonomial, polynomial ) );
            intRelation = add ( or ( op ( leq ), op ( geq ), op ( Op.EQUALS ) ),
                                sub ( add ( intF, nonNegMonomial), polynomial ) );
        }
        
        final Sort intS;

        final Function Z;                
        final Function add;        
        final Function mul;

        final Function leq;
        final Function geq;
        
        final TermFeature intF;
        
        final TermFeature addF;
        final TermFeature mulF;

        final TermFeature leqF;
        final TermFeature geqF;

        final TermFeature atom;
        
        // left-associatively arranged monomials
        final TermFeature monomial;
        // left-associatively arranged polynomials
        final TermFeature polynomial;
              
        final TermFeature literal;
        final TermFeature posLiteral;
        final TermFeature negLiteral;
        final TermFeature nonNegLiteral;
        final TermFeature nonPosLiteral;
        final TermFeature zeroLiteral;

        final TermFeature nonNegMonomial;
        final TermFeature negMonomial;
        final TermFeature nonCoeffMonomial;
        
        final TermFeature intEquation;
        final TermFeature intInEquation;
        final TermFeature intRelation;
    }

}
